@@ -5,6 +5,8 @@ import pickle
 import socket
 import numbers
 import functools
+import logging
+import unicodedata
 import subprocess
 from typing import Dict, List, Union
 from pathlib import Path
@@ -16,25 +18,26 @@ import msgpack_numpy as msgpack_np
 from beartype import beartype
 from mergedeep import Strategy, merge
 from typeguard import typechecked
-from beartype.cave import AnyType
+from beartype.cave import AnyType, NoneTypeOr
 import zsvision.zs_data_structures
 
 
 @functools.lru_cache(maxsize=64, typed=False)
-@typechecked
-def memcache(path: Union[Path, str]):
+@beartype
+def memcache(path: Union[Path, str], verbose: bool = True):
     path = Path(path)
     suffix = path.suffix
-    print(f"loading data from {path} ({socket.gethostname()})", end=" ", flush=True)
+    if verbose:
+        print(f"loading data from {path} ({socket.gethostname()})", end=" ", flush=True)
     tic = time.time()
-    if suffix in {".pkl", ".pickle"}:
-        res = pickle_loader(path)
+    if suffix in {".pkl", ".pickle", ".pckl"}:
+        res = pickle_loader(pkl_path=path, verbose=verbose)
     elif suffix in {".hkl", ".hickle"}:
         res = hickle.load(path)
     elif suffix == ".npy":
-        res = np_loader(path)
+        res = np_loader(path, verbose=verbose)
     elif suffix == ".mp":
-        res = msgpack_loader(path)
+        res = msgpack_loader(path, verbose=verbose)
     elif suffix == ".json":
         with open(path, "r") as f:
             res = json.load(f)
@@ -42,7 +45,8 @@ def memcache(path: Union[Path, str]):
         res = loadmat(path)
     else:
         raise ValueError(f"unknown suffix: {suffix} for path {path}")
-    print(f"[Total: {time.time() - tic:.1f}s]")
+    if verbose:
+        print(f"[Total: {time.time() - tic:.1f}s]")
     return res
 
 
@@ -58,7 +62,11 @@ def support_old_pickles(buffer: bytes) -> object:
 
 
 @beartype
-def pickle_loader(pkl_path: Path, backwards_compatible: bool = True) -> object:
+def pickle_loader(
+        pkl_path: Path,
+        verbose: bool,
+        backwards_compatible: bool = True,
+) -> object:
     """Deserialise object from pickle.
 
     Args:
@@ -83,27 +91,30 @@ def pickle_loader(pkl_path: Path, backwards_compatible: bool = True) -> object:
 
 
 @beartype
-def msgpack_loader(mp_path: Path):
+def msgpack_loader(mp_path: Path, verbose: bool):
     """Msgpack provides a faster serialisation routine than pickle, so is preferable
     for loading and deserialising large feature sets from disk."""
     tic = time.time()
     with open(mp_path, "rb") as f:
         buffer = f.read()
-        print(f"[I/O: {time.time() - tic:.1f}s]", end=" ")
+        if verbose:
+            print(f"[I/O: {time.time() - tic:.1f}s]", end=" ")
         tic = time.time()
         data = msgpack_np.unpackb(buffer, raw=False)
-        print(f"[deserialisation: {time.time() - tic:.1f}s]", end=" ")
+        if verbose:
+            print(f"[deserialisation: {time.time() - tic:.1f}s]", end=" ")
     return data
 
 
 @beartype
-def np_loader(np_path: Path, l2norm=False):
+def np_loader(np_path: Path, verbose: bool, l2norm: bool = False):
     with open(np_path, "rb") as f:
         data = np.load(f, encoding="latin1", allow_pickle=True)
     if isinstance(data, np.ndarray) and data.size == 1:
         data = data[()]  # handle numpy dict storage convnetion
     if l2norm:
-        print("L2 normalizing features")
+        if verbose:
+            print("L2 normalizing features")
         if isinstance(data, dict):
             for key in data:
                 feats_ = data[key]
@@ -268,15 +279,27 @@ def concat_features(feat_paths, axis):
 
 class BlockTimer:
     """A minimal inline codeblock timer"""
-    def __init__(self, msg, precise=False, mute=False):
+    @beartype
+    def __init__(
+            self,
+            msg: str,
+            mute: bool = False,
+            precise: bool = False,
+            logger: NoneTypeOr[logging.Logger] = False,
+    ):
         self.msg = msg
         self.mute = mute
         self.precise = precise
+        self.logger = logger
         self.start = None
 
     def __enter__(self):
         self.start = time.time()
-        print(f"{self.msg}...", end="", flush=True)
+        msg = f"{self.msg}..."
+        if self.logger:
+            self.logger.info(msg)
+        else:
+            print(msg, end="", flush=True)
         return self
 
     def __exit__(self, *args):
@@ -285,7 +308,11 @@ class BlockTimer:
         else:
             total = time.strftime('%Hh%Mm%Ss', time.gmtime(time.time() - self.start))
         if not self.mute:
-            print(f" took {total}")
+            msg = f" took {total}"
+            if self.logger:
+                self.logger.info(msg)
+            else:
+                print(msg)
 
 
 @typechecked
@@ -374,9 +401,9 @@ def list_visible_gpu_types() -> List[str]:
     return devices
 
 
-@typechecked
+@beartype
 def quote_and_escape_ffmpeg_path(path: (str, Path)) -> str:
-    """Quote paths for use with ffmpeg/ffprobe.
+    """Quote and escape paths for use with ffmpeg/ffprobe.
 
     Args:
         path: the location of a file to be processed by ffmpeg
@@ -384,13 +411,17 @@ def quote_and_escape_ffmpeg_path(path: (str, Path)) -> str:
     Returns:
         a quoted, dollar-escaped path
 
-    NOTE: This function is useful for processing file paths that may contain spaces or
-    dollar characters when invoking ffmpeg or ffprobe from python. Example usage for a
-    path to an input file:
+    Example usage:
         `os.system("ffprobe {quote_and_escape_ffmpeg_path(path)}")`
+
+    NOTE: This function is useful for processing file paths that may contain:
+        1. spaces
+        2. dollar characters ($)
+        3. percent sign characters (%)
+    when invoking ffmpeg or ffprobe from python.
     """
     # Dollar signs need to be escaped when used in paths
-    escaped = str(path).replace("$", r"\$")
+    escaped = str(path).replace("$", r"\$").replace("%", r"\%")
     if "'" in escaped:
         quoted = f'"{escaped}"'
     else:
@@ -398,5 +429,77 @@ def quote_and_escape_ffmpeg_path(path: (str, Path)) -> str:
     return quoted
 
 
+@beartype
+def parse_tree_layout(
+        tree_layout_path: Path,
+        prefix_token: str = "── ",
+) -> set:
+    """Given a text dump of the output of the linux `tree` command, this function will
+    reconstruct the relative paths of the files in the tree.
+
+    Args:
+        tree_layout_path: the location of the text file containing the `tree` output
+        prefix_token: the token used by the `tree` command to denote a new file.
+
+    Returns:
+        the collection of parsed paths.
+
+    NOTES:
+    1. This function assumes that it is parsing the output of the tree command that
+    has been run in the directory of the structure it is displaying (i.e. `tree` is run)
+    without arguments.
+    2. The output of each row in the `tree` command is prefixed by a T-bar or an L-bar
+    (see example formats (1) and (2) resp. below).
+    3. If the file at `tree_layout_path` contains any rows that are not part of the tree
+    output, they are ignored.
+
+    Example:
+        Given tree outputs of the forms (1) or (2) shown below:
+
+        (1)
+            ├── Conversation
+            │   ├── Belfast
+            │   │   ├── 11+12
+
+        (2)
+            └── Conversation
+                └── Belfast
+                    └── 11+12
+
+        in both cases, this function will return a set of pathlib paths of the form:
+
+        {
+            "."
+            "Conversation",
+            "Conversation/Belfast",
+            "Conversation/Belfast/11+12",
+        }
+
+    """
+    with open(tree_layout_path, "r") as f:
+        rows = f.read().splitlines()
+
+    # filter the input to only contain the file tree structure by searching for the
+    # presence of the tree prefix token
+    rows = [x for x in rows if prefix_token in x]
+
+    # convert nbsp escape codes into spaces
+    rows = [unicodedata.normalize("NFKD", x) for x in rows]
+
+    current_path = Path(".")
+    paths = {current_path}
+    known_prefix_heads = {"├", "└"}
+    for row in rows:
+        prefix, name = row.split(prefix_token)
+        prefix, prefix_head = prefix[:-1], list(prefix).pop(-1)
+        msg = f"Expected prefix head to be in {known_prefix_heads} found {prefix_head}"
+        assert prefix_head in known_prefix_heads, msg
+        assert len(prefix) % 4 == 0, "Expected prefix string length to be a multiple of 4"
+        depth = int(len(prefix) / 4)
+        current_path = Path(*current_path.parts[:depth]) / name
+        paths.add(current_path)
+    return paths
+
+
 if __name__ == "__main__":
     print(list_visible_gpu_types())
